from __future__ import annotations

import gmpy2

from .util import construct_random_polynomial, polynomial


class Player:

    """
    Player in a threshold ElGamal cryptosystem.

    :param i: Player identifier.
    :type i: int
    :param params: Parameters of the threshold scheme: primes p and q, generators g and h, total number of players n and threshold k.
    :type params: dict
    """

    def __init__(self, i, params):

        if not (isinstance(i, int) and isinstance(params, dict)):
            raise ValueError("Type mismatch")

        if {'g', 'h', 'q', 'p', 'n', 'k'} != set(params.keys()):
            raise ValueError("Parameters missing")

        if not all(isinstance(val, (gmpy2.mpz, int)) for val in params.values()):
            raise ValueError("Params type mismatch")

        self.id = i

        self.g, self.h = params['g'], params['h'],
        self.q, self.p = params['q'], params['p'],
        self.n, self.k = params['n'], params['k']
        self.t = self.k - 1

        self.a = construct_random_polynomial(self.t, self.q)
        self.b = construct_random_polynomial(self.t, self.q)
        self.f = lambda x: polynomial(self.a, x, self.q)
        self.f2 = lambda x: polynomial(self.b, x, self.q)

        self.pk = None
        self.received_shares = dict()
        self.received_shares2 = dict()
        self.decryption_share = None

        self.honest_players = None
        self.disqualified = False
        self.complaints_received = 0

    def __repr__(self):
        return f'Player({self.id})'

    def __str__(self):
        return f'Player {self.id}'

    def send_shares(self, players):
        """
        Sends a pair of shares to each of the players.

        :param players: List of players (including self).
        :type players: list
        """
        shares = [self.f(j) for j in range(1, self.n + 1)]
        shares2 = [self.f2(j) for j in range(1, self.n + 1)]
        for j, share, share2 in zip(players, shares, shares2):
            j.receive_share(self, share, share2)

    def commit(self):
        """
        Creates commitments for coefficient of the generated random polynomials.

        :return: Commitments.
        :rtype list:
        """
        commits = []
        for k in range(0, self.t+1):
            g_a = gmpy2.powmod(self.g, self.a[k], self.p)
            h_b = gmpy2.powmod(self.h, self.b[k], self.p)
            commit = gmpy2.mul(g_a, h_b)
            commit = gmpy2.mod(commit, self.p)
            commits.append(commit)
        return commits

    def commit_pk(self):
        """
        Generates commitments for the construction of the public key.

        :return: Commitments.
        :rtype: list
        """
        commits = []
        for k in range(0, self.t+1):
            g_a = gmpy2.powmod(self.g, self.a[k], self.p)
            commits.append(g_a)
        return commits

    def check_shares(self, commits, players):
        """
        Checks commitments for the shares and complains if the check fails.

        :param commits: Commitments.
        :type commits: dict
        :param players: List of players.
        :type players: list
        """
        if self.honest_players is None:
            self.honest_players = players
        for player in players:
            g_share = gmpy2.powmod(self.g, self.received_shares[player], self.p)
            h_share = gmpy2.powmod(self.h, self.received_shares2[player], self.p)
            check1 = gmpy2.mul(g_share, h_share)
            check1 = gmpy2.mod(check1, self.p)
            check2 = gmpy2.mpz(1)
            for k in range(0, self.t+1):
                exponent = self.id ** k
                factor = gmpy2.powmod(commits[player][k], exponent, self.p)
                check2 = gmpy2.mul(check2, factor)
                check2 = gmpy2.mod(check2, self.p)
            if check1 != check2:
                self.complain(player)

    def check_pk_commits(self, commits, failed_shares):
        """
        Checks commits for the construction of the public key. If any of the
        commits fail, the received share of that player is returned and used later
        when constructing the public key without the player's involvement.

        :param commits: Commitments.
        :type commits: dict
        :param failed_shares: Dictionary of received shares of players for whom the check failed. Used later while constructing the public key.
        :type failed_shares: dict
        """
        players = self.honest_players.copy()
        for i, player in enumerate(players):
            check1 = gmpy2.powmod(self.g, self.received_shares[player], self.p)
            product = gmpy2.mpz(1)
            for k in range(0, self.t+1):
                exponent = self.id ** k
                factor = gmpy2.powmod(commits[player][k], exponent, self.p)
                product = gmpy2.mul(product, factor)
                product = gmpy2.mod(product, self.p)
            check2 = gmpy2.mod(product, self.p)
            if check1 != check2:
                if player not in failed_shares:
                    failed_shares[player] = [(self.id, self.received_shares[player])]
                else:
                    temp = failed_shares[player]
                    temp.append((self.id, self.received_shares[player]))
                    failed_shares[player] = temp

    def get_pk(self):
        """
        Fetches the public key.

        :return: Public key.
        :rtype: gmpy2.mpz
        """
        if self.pk is None:
            raise ValueError("Public key not generated yet")
        else:
            return self.pk

    def construct_pk(self, commits):
        """
        Constructs public key from public key commitments.

        :param commits: Commitments.
        :type commits: dict
        """
        if self.pk is None:
            product = gmpy2.mpz(1)
            for i, pl in enumerate(self.honest_players):
                product = gmpy2.mul(commits[pl][0], product)
                product = gmpy2.mod(product, self.p)
            self.pk = product

    def construct_decryption_share(self, c1):
        """
        Constructs decryption share.

        :param c1: c1
        :type c1: gmpy2.mpz
        """
        if self.decryption_share is None:
            sum_ = gmpy2.mpz()
            for player, share in self.received_shares.items():
                if player not in self.honest_players:
                    continue
                sum_ = gmpy2.add(sum_, share)
                sum_ = gmpy2.mod(sum_, self.q)
            d_share = gmpy2.powmod(c1, sum_, self.p)
            self.decryption_share = d_share

    def get_decryption_share(self, c1):
        """
        Fetches the decryption share.

        :param c1: c1
        :type c1: gmpy2.mpz
        :return: Decryption share.
        :rtype: gmpy2.mpz
        """
        if self.decryption_share is None:
            self.construct_decryption_share(c1)
            return self.decryption_share
        else:
            return self.decryption_share

    def receive_share(self, player, share, share2):
        self.received_shares[player] = share
        self.received_shares2[player] = share2

    def complain(self, player):
        """
        Processes complains to player. If the player received more than t
        complaints, they are disqualified.

        :param player: Player.
        :type player: Player
        """
        player.complaints_received += 1
        if player.complaints_received > self.t:
            player.disqualified = True
            self.honest_players.remove(player)



import time

from threshold_elgamal import *


def time_func(func, *args, **kwargs):
    runtimes = 0
    for _ in range(3):
        start_time = time.perf_counter()
        func(*args, **kwargs)
        end_time = time.perf_counter()
        runtime = end_time - start_time
        runtimes += runtime
    avg_runtime_ = runtimes / 3
    return avg_runtime_


for length in [2048, 3072, 4096]:
    for n in [1, 3, 5, 10, 20]:
        for k in [1, int(n/2), n-1]:
            try:
                avg_runtime = time_func(run_tc_scheme, k, n, 10, length)
                print(length, n, k, avg_runtime, sep=" & ")
            except BaseException:
                continue




import random

from threshold_elgamal import run_tc_scheme, create_tc_scheme, Player, ThresholdElGamal
from threshold_elgamal.util import *
import unittest


class TCTestCase(unittest.TestCase):

    def setUp(self) -> None:
        pass

    def test_simple(self):
        res = run_tc_scheme(3, 5, 10)
        self.assertTrue(res)

    def test_bigger(self):
        res = run_tc_scheme(5, 7, 10)
        self.assertTrue(res)

    def test_1_of_n_scheme(self):
        res = run_tc_scheme(1, 5, 10)
        self.assertTrue(res)

    def test_n_of_n_scheme(self):
        res = run_tc_scheme(5, 5, 10)
        self.assertTrue(res)

    def test_1_of_1_scheme(self):
        res = run_tc_scheme(1, 1, 10)
        self.assertTrue(res)

    def test_incorrect_scheme(self):
        self.assertRaises(ValueError, run_tc_scheme, 6, 5, 10)

    def test_incorrect_scheme_zero(self):
        self.assertRaises(ValueError, run_tc_scheme, 0, 6, 10)

    def test_incorrect_msg(self):
        self.assertRaises(ValueError, run_tc_scheme, 5, 6, "message")

    def test_msg_too_large(self):
        y, _, scheme = create_tc_scheme(3, 5)
        message = scheme.p + 50
        self.assertRaises(ValueError, scheme.encrypt, y, message)

    def test_3072_scheme(self):
        res = run_tc_scheme(3, 5, 10, 3072)
        self.assertTrue(res)

    def test_4096_scheme(self):
        res = run_tc_scheme(3, 5, 10, 4096)
        self.assertTrue(res)


class ManualTCTestCase(unittest.TestCase):

    def test_manual_scheme(self):
        y, players, scheme = create_tc_scheme(3, 5)
        c1, c2 = scheme.encrypt(y, 10)
        decryption_shares = {pl.id: pl.get_decryption_share(c1) for pl in players}
        dec_msg = scheme.decrypt(c2, decryption_shares)
        self.assertTrue(10 == dec_msg)

    def test_manual_1_of_n_scheme(self):
        y, players, scheme = create_tc_scheme(1, 5)
        c1, c2 = scheme.encrypt(y, 10)
        decryption_shares = {pl.id: pl.get_decryption_share(c1) for pl in players}
        dec_msg = scheme.decrypt(c2, decryption_shares)
        self.assertTrue(10 == dec_msg)

    def test_manual_1_of_1_scheme(self):
        y, players, scheme = create_tc_scheme(1, 1)
        c1, c2 = scheme.encrypt(y, 10)
        decryption_shares = {pl.id: pl.get_decryption_share(c1) for pl in players}
        dec_msg = scheme.decrypt(c2, decryption_shares)
        self.assertTrue(10 == dec_msg)

    def test_manual_n_of_n_scheme(self):
        y, players, scheme = create_tc_scheme(5, 5)
        c1, c2 = scheme.encrypt(y, 10)
        decryption_shares = {pl.id: pl.get_decryption_share(c1) for pl in players}
        dec_msg = scheme.decrypt(c2, decryption_shares)
        self.assertTrue(10 == dec_msg)

    def test_manual_incorrect_scheme(self):
        y, players, scheme = create_tc_scheme(3, 5)
        c1, c2 = scheme.encrypt(y, 10)
        decryption_shares = {pl.id: pl.get_decryption_share(c1) for pl in players}
        new_shares = random.sample(decryption_shares.keys(), 2)
        new_shares = {idx: decryption_shares[idx] for idx in new_shares}
        self.assertRaises(ValueError, scheme.decrypt, c2, new_shares)

    def test_manual_incorrect_scheme2(self):
        y, players, scheme = create_tc_scheme(3, 5)
        c1, c2 = scheme.encrypt(y, 10)
        self.assertRaises(ValueError, scheme.decrypt, c2, [])


class PlayerTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.scheme = ThresholdElGamal()
        self.params = self.scheme.get_or_generate_params()
        self.missing_params = self.params.copy()
        del self.missing_params["k"]
        self.incorrect_params = self.missing_params.copy()
        self.incorrect_params["k"] = "string"

    def test_player_creation_string(self):
        self.assertRaises(ValueError, Player, "string", self.params)

    def test_player_creation_empty_list(self):
        self.assertRaises(ValueError, Player, 1, dict())

    def test_player_creation_missing_param(self):
        self.assertRaises(ValueError, Player, 1, self.missing_params)

    def test_player_creation_incorrect_param(self):
        self.assertRaises(ValueError, Player, 1, self.incorrect_params)


class DishonestPlayerTestCase(unittest.TestCase):

    def setUp(self) -> None:
        self.scheme = ThresholdElGamal(3, 5)
        self.params = self.scheme.get_or_generate_params()

    def test_one_dishonest_player(self):
        players = [Player(i, self.params) for i in range(1, self.scheme.n + 1)]
        commits = {pl: pl.commit() for pl in players}
        misfit = players[random.randint(0, self.scheme.n-1)]
        misfit.a = construct_random_polynomial(self.scheme.k - 1, self.scheme.q)

        for pl in players:
            pl.send_shares(players)

        for pl in players.copy():
            pl.check_shares(commits, players)

        self.assertNotIn(misfit, players)

    def test_two_dishonest_players(self):
        players = [Player(i, self.params) for i in range(1, self.scheme.n + 1)]
        commits = {pl: pl.commit() for pl in players}

        misfit1 = players[random.randint(0, self.scheme.n-1)]
        misfit1.a = construct_random_polynomial(self.scheme.k - 1, self.scheme.q)
        misfit2 = players[(misfit1.id + 1) % self.scheme.n]
        misfit2.a = construct_random_polynomial(self.scheme.k - 1, self.scheme.q)

        for pl in players:
            pl.send_shares(players)

        for pl in players.copy():
            pl.check_shares(commits, players)

        self.assertNotIn(misfit1, players)
        self.assertNotIn(misfit2, players)

    def test_t_dishonest_players(self):
        players = [Player(i, self.params) for i in range(1, self.scheme.n + 1)]
        commits = {pl: pl.commit() for pl in players}

        misfits = random.sample(players, self.scheme.k - 1)
        for m in misfits:
            m.a = construct_random_polynomial(self.scheme.k - 1, self.scheme.q)

        for pl in players:
            pl.send_shares(players)

        for pl in players.copy():
            pl.check_shares(commits, players)

        for m in misfits:
            self.assertNotIn(m, players)

    def test_dishonest_reconstruction(self):

        players = [Player(i, self.params) for i in range(1, self.scheme.n + 1)]
        commits = {pl: pl.commit() for pl in players}

        for pl in players:
            pl.send_shares(players)

        for pl in players.copy():
            pl.check_shares(commits, players)

        misfit = random.sample(players, 1)[0]
        old_polynomial = misfit.a
        misfit.a = construct_random_polynomial(self.scheme.k - 1, self.scheme.q)

        commits_pk = {pl: pl.commit_pk() for pl in players}
        failed_shares = dict()
        for pl in players:
            pl.check_pk_commits(commits_pk, failed_shares)

        recalculated_commits = dict()
        for i, points in failed_shares.items():
            z = reconstruct_polynomial(points, self.scheme.k, self.scheme.q)
            self.assertEqual(z, old_polynomial[0])
            recalculated_commits[i] = gmpy2.powmod(self.scheme.g, z, self.scheme.p)

    def test_dishonest_pk_shares(self):

        players = [Player(i, self.params) for i in range(1, self.scheme.n + 1)]
        commits = {pl: pl.commit() for pl in players}

        for pl in players:
            pl.send_shares(players)

        for pl in players.copy():
            pl.check_shares(commits, players)

        misfit = random.sample(players, 1)[0]
        misfit.a = construct_random_polynomial(self.scheme.k - 1, self.scheme.q)

        commits_pk = {pl: pl.commit_pk() for pl in players}
        failed_shares = dict()
        for pl in players:
            pl.check_pk_commits(commits_pk, failed_shares)

        recalculated_commits = dict()
        for pl, points in failed_shares.items():
            z = reconstruct_polynomial(points, self.scheme.k, self.scheme.q)
            recalculated_commits[pl] = gmpy2.powmod(self.scheme.g, z, self.scheme.p)

        for pl in recalculated_commits.keys():
            commits_pk[pl] = [recalculated_commits[pl]]

        for pl in players:
            pl.construct_pk(commits_pk)

        y = players[0].get_pk()

        c1, c2 = self.scheme.encrypt(y, 10)
        decryption_shares = {pl.id: pl.get_decryption_share(c1) for pl in players}
        dec_msg = self.scheme.decrypt(c2, decryption_shares)
        self.assertEqual(dec_msg, 10)

    def test_t_dishonest_pk_shares(self):

        players = [Player(i, self.params) for i in range(1, self.scheme.n + 1)]
        commits = {pl: pl.commit() for pl in players}

        for pl in players:
            pl.send_shares(players)

        for pl in players.copy():
            pl.check_shares(commits, players)

        misfits = random.sample(players, self.scheme.k - 1)
        for misfit in misfits:
            misfit.a = construct_random_polynomial(self.scheme.k - 1, self.scheme.q)

        commits_pk = {pl: pl.commit_pk() for pl in players}
        failed_shares = dict()
        for pl in players:
            pl.check_pk_commits(commits_pk, failed_shares)

        recalculated_commits = dict()
        for pl, points in failed_shares.items():
            z = reconstruct_polynomial(points, self.scheme.k - 1, self.scheme.q)
            recalculated_commits[pl] = gmpy2.powmod(self.scheme.g, z, self.scheme.p)

        for pl in recalculated_commits.keys():
            commits_pk[pl] = [recalculated_commits[pl]]

        for pl in players:
            pl.construct_pk(commits_pk)

        y = players[0].get_pk()

        c1, c2 = self.scheme.encrypt(y, 10)
        decryption_shares = {pl.id: pl.get_decryption_share(c1) for pl in players}
        dec_msg = self.scheme.decrypt(c2, decryption_shares)
        self.assertEqual(dec_msg, 10)

    def test_dishonest_player_and_pk_shares(self):
        players = [Player(i, self.params) for i in range(1, self.scheme.n + 1)]
        # commits = [pl.commit() for pl in players]
        commits = {pl: pl.commit() for pl in players}

        misfits = random.sample(players, self.scheme.k - 1)
        for m in misfits:
            m.a = construct_random_polynomial(self.scheme.k - 1, self.scheme.q)

        for pl in players:
            pl.send_shares(players)

        for pl in players.copy():
            pl.check_shares(commits, players)

        for m in misfits:
            self.assertNotIn(m, players)

        misfits = random.sample(players, 1)
        for misfit in misfits:
            misfit.a = construct_random_polynomial(self.scheme.k - 1, self.scheme.q)

        commits_pk = {pl: pl.commit_pk() for pl in players}
        failed_shares = dict()
        for pl in players:
            pl.check_pk_commits(commits_pk, failed_shares)

        recalculated_commits = dict()
        for i, points in failed_shares.items():
            z = reconstruct_polynomial(points, self.scheme.k - 1, self.scheme.q)
            recalculated_commits[i] = gmpy2.powmod(self.scheme.g, z, self.scheme.p)

        for i in recalculated_commits.keys():
            commits_pk[i] = [recalculated_commits[i]]

        for pl in players:
            pl.construct_pk(commits_pk)

        y = players[0].get_pk()

        c1, c2 = self.scheme.encrypt(y, 10)
        decryption_shares = {pl.id: pl.get_decryption_share(c1) for pl in players}
        dec_msg = self.scheme.decrypt(c2, decryption_shares)
        self.assertEqual(dec_msg, 10)


class UtilTestCase(unittest.TestCase):

    def test_reconstruct_polynomial(self):
        coeffs = construct_random_polynomial(3, 499)
        points = [(i, polynomial(coeffs, i, 499)) for i in range(1, 8)]
        z = reconstruct_polynomial(points, 3, 499)
        self.assertEqual(z, coeffs[0])

    def test_reconstruct_polynomial_big(self):
        coeffs = construct_random_polynomial(10, 499)
        points = [(i, polynomial(coeffs, i, 499)) for i in range(1, 20)]
        z = reconstruct_polynomial(points, 10, 499)
        self.assertEqual(z, coeffs[0])


def suite():
    suite_ = unittest.TestSuite()
    suite_.addTests([TCTestCase(), ManualTCTestCase(), PlayerTestCase(), UtilTestCase(), DishonestPlayerTestCase()])
    return suite_


if __name__ == "__main__":
    unittest.TextTestRunner().run(suite())



from setuptools import setup
from codecs import open

setup(
    name='threshold_elgamal',
    version='0.0.8',
    long_description_content_type="text/markdown",
    long_description=open('README.md', 'r').read(),
    packages=['tests', 'threshold_elgamal'],
    url='',
    license='',
    author='Katarina Boras',
    author_email='katarina.boras2@outlook.com',
    description='Threshold ElGamal cryptosystem.'
)



import cProfile
from threshold_elgamal import run_tc_scheme
from pstats import Stats, SortKey

cProfile.run('run_tc_scheme(3, 5, 10)', filename='stats.txt')
p = Stats('stats.txt')
p.strip_dirs().sort_stats(SortKey.CALLS).print_stats()
